"""Simple IRC line parser.

Backported from async-irc (https://github.com/snoonetIRC/async-irc.git)
"""

import datetime
import re
import warnings
from abc import ABCMeta, abstractmethod
from collections.abc import Iterable, Iterator, Sequence
from typing import Final, Literal, Optional, TypeAlias, cast

from typing_extensions import Self

from irclib.errors import ParseError

__all__ = (
    "Cap",
    "CapList",
    "MessageTag",
    "TagList",
    "Prefix",
    "ParamList",
    "Message",
)
MsgTagList: TypeAlias = Optional["TagList"]
MsgPrefix: TypeAlias = Optional["Prefix"]
MessageTuple: TypeAlias = tuple[MsgTagList, MsgPrefix, str, "ParamList"]

TAGS_SENTINEL: Final = "@"
TAGS_SEP: Final = ";"
TAG_VALUE_SEP: Final = "="

PREFIX_SENTINEL: Final = ":"
PREFIX_USER_SEP: Final = "!"
PREFIX_HOST_SEP: Final = "@"

PARAM_SEP: Final = " "
TRAIL_SENTINEL: Final = ":"

CAP_SEP: Final = " "
CAP_VALUE_SEP: Final = "="

PREFIX_RE: Final = re.compile(
    r"^:?(?P<nick>.*?)(?:!(?P<user>.*?))?(?:@(?P<host>.*?))?$"
)

TAG_VALUE_ESCAPES: Final = {
    "\\s": " ",
    "\\:": ";",
    "\\r": "\r",
    "\\n": "\n",
    "\\\\": "\\",
}

TAG_VALUE_UNESCAPES: Final = {
    unescaped: escaped for escaped, unescaped in TAG_VALUE_ESCAPES.items()
}


def parse_server_time(
    value: str | None, default: datetime.datetime
) -> datetime.datetime:
    if value:
        return datetime.datetime.strptime(
            value, "%Y-%m-%dT%H:%M:%S.%fZ"
        ).replace(tzinfo=datetime.timezone.utc)

    return default


class Parseable(metaclass=ABCMeta):
    """Abstract class for parseable objects."""

    @classmethod
    @abstractmethod
    def parse(cls, text: str) -> Self:
        """Parse the object from a string."""
        raise NotImplementedError

    @abstractmethod
    def __str__(self) -> str:
        raise NotImplementedError


class Cap(Parseable):
    """Represents a CAP entity as defined in IRCv3.2."""

    def __init__(self, name: str, value: str | None = None) -> None:
        """Construct Cap object.

        Args:
            name: Cap name
            value: Optional Cap value. Defaults to None.
        """
        self._name = name
        self._value = value

    @property
    def name(self) -> str:
        """CAP name."""
        return self._name

    @property
    def value(self) -> str | None:
        """CAP value."""
        return self._value

    def as_tuple(self) -> tuple[str, str | None]:
        """Get data as a tuple of values."""
        return self.name, self.value

    @classmethod
    def parse(cls, text: str) -> Self:
        """Parse a CAP entity from a string."""
        name, _, value = text.partition(CAP_VALUE_SEP)
        return cls(name, value or None)

    def __eq__(self, other: object) -> bool:
        """Compare against another cap."""
        if isinstance(other, str):
            return self == self.parse(other)

        if isinstance(other, Cap):
            return self.as_tuple() == other.as_tuple()

        return NotImplemented

    def __ne__(self, other: object) -> bool:
        """Compare against another cap."""
        if isinstance(other, str):
            return self != self.parse(other)

        if isinstance(other, Cap):
            return self.as_tuple() != other.as_tuple()

        return NotImplemented

    def __str__(self) -> str:
        """Represent cap as a string."""
        if self.value:
            return CAP_VALUE_SEP.join((self.name, self.value))

        return self.name


class CapList(Parseable, list[Cap]):
    """Represents a list of CAP entities."""

    @classmethod
    def parse(cls, text: str) -> Self:
        """Parse a list of CAPs from a string."""
        if text.startswith(":"):
            text = text[1:]  # Remove leading colon

        # We want to strip any leading or trailing whitespace
        # Some networks (ie: freenode) send a trailing space in a CAP ACK
        stripped = text.strip()

        caps: Iterable[Cap]
        caps = (
            [] if not text else (Cap.parse(s) for s in stripped.split(CAP_SEP))
        )

        return cls(caps)

    def __eq__(self, other: object) -> bool:
        """Compare to a string cap list or list of Cap objects."""
        if isinstance(other, str):
            return self == self.parse(other)

        if isinstance(other, list):
            return list(self) == list(other)

        return NotImplemented

    def __ne__(self, other: object) -> bool:
        """Compare to a string cap list or sequence of Cap objects."""
        if isinstance(other, str):
            return self != self.parse(other)

        if isinstance(other, list):
            return list(self) != list(other)

        return NotImplemented

    def __str__(self) -> str:
        """Represent the list of caps as a string."""
        return CAP_SEP.join(map(str, self))


class MessageTag(Parseable):
    """Basic class to wrap a message tag."""

    def __init__(
        self, name: str, value: str = "", *, has_value: bool = False
    ) -> None:
        """Construct a message tag object.

        Args:
            name: Tag name
            value: Optional tag value. Defaults to "".
            has_value: Whether this tag has a value set. Defaults to False.
        """
        self._name = name
        self._value = value or ""
        self._has_value = has_value

    @property
    def name(self) -> str:
        """Message tag name."""
        return self._name

    @property
    def value(self) -> str:
        """Message tag value."""
        return self._value

    @staticmethod
    def unescape(value: str) -> str:
        """Replace the escaped characters in a tag value with their literals.

        :param value: Escaped string
        :return: Unescaped string
        """
        new_value = ""
        escaped = False
        for char in value:
            if escaped:
                new_value += TAG_VALUE_ESCAPES.get(f"\\{char}", char)
                escaped = False
            elif char == "\\":
                escaped = True
            else:
                new_value += char

        return new_value

    @staticmethod
    def escape(value: str) -> str:
        """Replace characters with their escaped variants.

        :param value: The raw string
        :return: The escaped string
        """
        return "".join(TAG_VALUE_UNESCAPES.get(c, c) for c in value)

    @classmethod
    def parse(cls, text: str) -> Self:
        """Parse a tag from a string.

        Arguments:
            text: The basic tag string

        Returns:
            The MessageTag object
        """
        name, sep, value = text.partition(TAG_VALUE_SEP)
        if value:
            value = MessageTag.unescape(value)

        return cls(name, value, has_value=bool(sep))

    def __eq__(self, other: object) -> bool:
        """Compare tag to a string representing a tag or another tag object."""
        if isinstance(other, str):
            return self == self.parse(other)

        if isinstance(other, MessageTag):
            return self.name == other.name and self.value == other.value

        return NotImplemented

    def __ne__(self, other: object) -> bool:
        """Compare tag to a string representing a tag or another tag object."""
        if isinstance(other, str):
            return self != self.parse(other)

        if isinstance(other, MessageTag):
            return not (self.name == other.name and self.value == other.value)

        return NotImplemented

    def __repr__(self) -> str:
        """Represent the tag object in a form useful for debugging."""
        return f"MessageTag(name={self.name!r}, value={self.value!r})"

    def __str__(self) -> str:
        """Represent the tag object as a string representation of a tag."""
        if self.value or self._has_value:
            return f"{self.name}{TAG_VALUE_SEP}{self.escape(self.value)}"

        return self.name


class TagList(Parseable, dict[str, MessageTag]):
    """Object representing the list of message tags on a line."""

    def __init__(self, tags: Iterable[MessageTag] = ()) -> None:
        """Construct a tag list with optional tags.

        Args:
            tags: Tags to initialize the list with. Defaults to ().
        """
        super().__init__((tag.name, tag) for tag in tags)

    @staticmethod
    def _cmp_type_map(
        obj: object,
    ) -> (
        tuple[dict[str, MessageTag], Literal[True]]
        | tuple[None, Literal[False]]
    ):
        if isinstance(obj, str):
            return TagList.parse(obj), True

        if isinstance(obj, dict):
            sample = next(iter(obj.values()), None)
            if obj and (sample is None or isinstance(sample, str)):
                # Handle str -> str dict
                return TagList.from_dict(obj), True

            # Handle str -> MessageTag dict
            return dict(obj), True

        if isinstance(obj, list):
            return TagList(obj), True

        return None, False

    @classmethod
    def parse(cls, text: str) -> Self:
        """Parse the list of tags from a string.

        :param text: The string to parse
        :return: The parsed object
        """
        return cls(map(MessageTag.parse, filter(None, text.split(TAGS_SEP))))

    @classmethod
    def from_dict(cls, tags: dict[str, str]) -> Self:
        """Create a TagList from a dict of tags."""
        return cls(MessageTag(k, v) for k, v in tags.items())

    def __eq__(self, other: object) -> bool:
        """Compare to another tag list, string, or list of MessageTag objects."""
        res = self._cmp_type_map(other)
        if not res[1]:
            return NotImplemented

        return dict(self) == dict(res[0])

    def __ne__(self, other: object) -> bool:
        """Compare to another tag list, string, or list of MessageTag objects."""
        res = self._cmp_type_map(other)
        if not res[1]:
            return NotImplemented

        return dict(self) != dict(res[0])

    def __str__(self) -> str:
        """Represent the tag list as a string."""
        return TAGS_SEP.join(map(str, self.values()))


class Prefix(Parseable):
    """Object representing the prefix of a line."""

    def __init__(
        self,
        nick: str | None = None,
        user: str | None = None,
        host: str | None = None,
    ) -> None:
        """Construct a prefix."""
        self._nick = nick or ""
        self._user = user or ""
        self._host = host or ""

    @property
    def nick(self) -> str:
        """IRC nickname."""
        return self._nick

    @property
    def user(self) -> str:
        """IRC ident/username."""
        return self._user

    @property
    def ident(self) -> str:
        """IRC ident. Alias for `user`."""
        return self.user

    @property
    def host(self) -> str:
        """Hostname."""
        return self._host

    @property
    def mask(self) -> str:
        """The complete n!u@h mask."""
        mask = self.nick
        if self.user:
            mask += PREFIX_USER_SEP + self.user

        if self.host:
            mask += PREFIX_HOST_SEP + self.host

        return mask

    @property
    def _data(self) -> tuple[str, str, str]:
        return self.nick, self.user, self.host

    @classmethod
    def parse(cls, text: str) -> Self:
        """Parse the prefix from a string.

        Arguments:
            text: String to parse

        Returns:
            Parsed Object
        """
        if not text:
            return cls()

        match = PREFIX_RE.match(text)
        if not match:  # pragma: no cover
            # This should never trip, we are pretty lenient with prefixes
            msg = "Invalid IRC prefix format"
            raise ParseError(msg)

        nick, user, host = match.groups()
        return cls(nick, user, host)

    def __iter__(self) -> Iterator[str]:
        """Iterator over the prefix components."""
        return iter(self._data)

    def __eq__(self, other: object) -> bool:
        """Compare to prefix string or another prefix object."""
        if isinstance(other, str):
            return self == self.parse(other)

        if isinstance(other, Prefix):
            return self._data == other._data

        return NotImplemented

    def __ne__(self, other: object) -> bool:
        """Compare to prefix string or another prefix object."""
        if isinstance(other, str):
            return self != self.parse(other)

        if isinstance(other, Prefix):
            return self._data != other._data

        return NotImplemented

    def __bool__(self) -> bool:
        """Return whether the prefix contains any values."""
        return any(self)

    def __str__(self) -> str:
        """Represent the prefix as a string."""
        return self.mask


class ParamList(Parseable, list[str]):
    """An object representing the parameter list from a line."""

    def __init__(self, *params: str, has_trail: bool = False) -> None:
        """Construct list of parameters."""
        super().__init__(params)
        self._has_trail = has_trail

    @property
    def has_trail(self) -> bool:
        """Does the parameter list end with a trailing parameter."""
        return self._has_trail

    @classmethod
    def from_list(cls, data: Sequence[str]) -> Self:
        """Create a ParamList from a Sequence of strings."""
        if not data:
            return cls()

        args = list(data[:-1])
        if data[-1].startswith(TRAIL_SENTINEL) or not data[-1]:
            has_trail = True
            args.append(data[-1])
        else:
            has_trail = False
            args.append(data[-1])

        return cls(*args, has_trail=has_trail)

    @classmethod
    def parse(cls, text: str) -> Self:
        """Parse a list of parameters.

        Arguments:
            text: The list of parameters

        Returns:
            The parsed object
        """
        args = []
        has_trail = False
        while text:
            if text[0] == TRAIL_SENTINEL:
                args.append(text[1:])
                has_trail = True
                break

            arg, _, text = text.partition(PARAM_SEP)
            if arg:
                args.append(arg)

        return cls(*args, has_trail=has_trail)

    def __eq__(self, other: object) -> bool:
        """Compare to string or list of parameters."""
        if isinstance(other, str):
            return self == self.parse(other)

        if isinstance(other, list):
            return list(self) == list(self.from_list(other))

        return NotImplemented

    def __ne__(self, other: object) -> bool:
        """Compare to string or list of parameters."""
        if isinstance(other, str):
            return self != self.parse(other)

        if isinstance(other, list):
            return list(self) != list(other)

        return NotImplemented

    def __str__(self) -> str:
        """Represent this parameter list as a string."""
        if not self:
            return ""

        needs_trail = (
            PARAM_SEP in self[-1]
            or self[-1].startswith(TRAIL_SENTINEL)
            or not self[-1]
        )

        if self.has_trail or needs_trail:
            return PARAM_SEP.join(self[:-1] + [TRAIL_SENTINEL + self[-1]])

        return PARAM_SEP.join(self)


def _parse_tags(
    tags: TagList | dict[str, str] | str | None | list[str],
) -> MsgTagList:
    if isinstance(tags, TagList):
        return tags

    if isinstance(tags, dict):
        return TagList.from_dict(cast(dict[str, str], tags))

    if isinstance(tags, str):
        return TagList.parse(tags)

    if tags is None:
        return None

    return TagList(MessageTag.parse(str(tag)) for tag in tags)


def _parse_prefix(prefix: Prefix | str | None | Iterable[str]) -> MsgPrefix:
    if isinstance(prefix, Prefix):
        return prefix

    if isinstance(prefix, str):
        return Prefix.parse(prefix)

    if prefix is None:
        return None

    return Prefix(*prefix)


def _parse_params(
    parameters: tuple[str | list[str] | ParamList, ...],
) -> ParamList:
    if len(parameters) == 1 and not isinstance(parameters[0], str):
        # This seems to be a list
        if isinstance(parameters[0], ParamList):
            return parameters[0]

        return ParamList.from_list(parameters[0])

    return ParamList.from_list(cast(tuple[str, ...], parameters))


class Message(Parseable):
    """An object representing a parsed IRC line."""

    def __init__(
        self,
        tags: TagList | dict[str, str] | str | None | list[str],
        prefix: str | Prefix | None | Iterable[str],
        command: str,
        *parameters: str | list[str] | ParamList,
        time: datetime.datetime | None = None,
    ) -> None:
        """Construct message object."""
        self._tags = _parse_tags(tags)
        self._prefix = _parse_prefix(prefix)
        self._command = command
        self._parameters = _parse_params(parameters)

        if time is None:
            time = datetime.datetime.now(datetime.timezone.utc)

        if time.tzinfo is None:
            time = time.replace(tzinfo=datetime.timezone.utc)
            warnings.warn(
                "Timezone-naive datetime is deprecated for 'time'.",
                DeprecationWarning,
                stacklevel=2,
            )

        self._time = time

    def has_tag(self, name: str) -> bool:
        """Return whether this message has a particular message tag."""
        if not self.tags:
            return False

        return name in self.tags

    def get_tag_value(self, name: str) -> str | None:
        """Get value for a message tag, or None if not set."""
        if self.tags and name in self.tags:
            return self.tags[name].value

        return None

    @property
    def time(self) -> datetime.datetime:
        """Time the message was sent/received.

        Uses server-time tag if available.
        """
        return parse_server_time(self.get_tag_value("time"), self._time)

    @property
    def message_id(self) -> str | None:
        """Unique message ID provided by server."""
        return self.get_tag_value("msgid")

    @property
    def batch_id(self) -> str | None:
        """Batch ID provided by server."""
        return self.get_tag_value("batch")

    @property
    def tags(self) -> MsgTagList:
        """IRC tag list."""
        return self._tags

    @property
    def prefix(self) -> MsgPrefix:
        """IRC prefix."""
        return self._prefix

    @property
    def command(self) -> str:
        """IRC command."""
        return self._command

    @property
    def parameters(self) -> ParamList:
        """Command parameter list."""
        return self._parameters

    def as_tuple(self) -> MessageTuple:
        """Get the message object as a tuple of values."""
        return self.tags, self.prefix, self.command, self.parameters

    @classmethod
    def parse(
        cls, text: str | bytes, *, time: datetime.datetime | None = None
    ) -> Self:
        """Parse an IRC message in to objects."""
        if isinstance(text, memoryview):
            text = text.tobytes().decode(errors="ignore")

        if isinstance(text, bytes | bytearray):
            text = text.decode(errors="ignore")

        tags = ""
        prefix = ""
        if text.startswith(TAGS_SENTINEL):
            tags, _, text = text.partition(PARAM_SEP)

        if text.startswith(PREFIX_SENTINEL):
            prefix, _, text = text.partition(PARAM_SEP)

        command, _, params = text.partition(PARAM_SEP)
        # Differentiate empty tags '@ CMD' from no tags 'CMD'
        tags_obj = TagList.parse(tags[1:]) if tags else None
        # Differentiate empty prefix ': CMD' from no prefix 'CMD'
        prefix_obj = Prefix.parse(prefix[1:]) if prefix else None
        command = command.upper()
        param_obj = ParamList.parse(params)
        return cls(tags_obj, prefix_obj, command, param_obj, time=time)

    def __eq__(self, other: object) -> bool:
        """Compare to another message which can be str, bytes, or a Message object."""
        if isinstance(other, str | bytes):
            return self == Message.parse(other)

        if isinstance(other, Message):
            return self.as_tuple() == other.as_tuple()

        return NotImplemented

    def __ne__(self, other: object) -> bool:
        """Compare to another message which can be str, bytes, or a Message object."""
        if isinstance(other, str | bytes):
            return self != Message.parse(other)

        if isinstance(other, Message):
            return self.as_tuple() != other.as_tuple()

        return NotImplemented

    def __bool__(self) -> bool:
        """Check if the line is empty or not."""
        return any(self.as_tuple())

    def __str__(self) -> str:
        """Represent this Message as a properly formatted IRC line."""
        tag_str = "" if self.tags is None else TAGS_SENTINEL + str(self.tags)
        prefix_str = (
            "" if self.prefix is None else PREFIX_SENTINEL + str(self.prefix)
        )

        return PARAM_SEP.join(
            str(s)
            for s in (tag_str, prefix_str, self.command, self.parameters)
            if s
        )



"""IRC numeric mapping."""

from collections.abc import Iterator, Mapping
from dataclasses import dataclass

__all__ = ("Numeric", "numerics")


@dataclass
class Numeric:
    """IRC numeric information."""

    name: str
    numeric: int


class NumericsDict(Mapping[str, Numeric]):
    """Mapping of IRC numerics."""

    def __init__(self, *args: Numeric) -> None:
        super().__init__()
        self.data = args
        self.names = {num.name: num for num in args}
        self.nums = {num.numeric: num for num in args}
        self.num_strs = {str(num.numeric).zfill(3): num for num in args}

    def from_int(self, n: int) -> Numeric:
        """Get a numeric by its number."""
        return self.nums[n]

    def from_name(self, name: str) -> Numeric:
        """Get a numeric by name."""
        return self.names[name.upper()]

    def __getattr__(self, item: str) -> Numeric:
        try:
            return self.names[str(item).upper()]
        except LookupError:
            raise AttributeError(item) from None

    def __getitem__(self, key: str) -> Numeric:
        return self.nums[int(key)]

    def __iter__(self) -> Iterator[str]:
        return iter(self.num_strs)

    def __len__(self) -> int:
        return len(self.names)


numerics = NumericsDict(
    Numeric("RPL_WELCOME", 1),
    Numeric("RPL_YOURHOST", 2),
    Numeric("RPL_CREATED", 3),
    Numeric("RPL_MYINFO", 4),
    Numeric("RPL_BOUNCE", 5),
    Numeric("RPL_TRACELINK", 200),
    Numeric("RPL_TRACECONNECTING", 201),
    Numeric("RPL_TRACEHANDSHAKE", 202),
    Numeric("RPL_TRACEUNKNOWN", 203),
    Numeric("RPL_TRACEOPERATOR", 204),
    Numeric("RPL_TRACEUSER", 205),
    Numeric("RPL_TRACESERVER", 206),
    Numeric("RPL_TRACESERVICE", 207),
    Numeric("RPL_TRACENEWTYPE", 208),
    Numeric("RPL_TRACECLASS", 209),
    Numeric("RPL_TRACERECONNECT", 210),
    Numeric("RPL_STATSLINKINFO", 211),
    Numeric("RPL_STATSCOMMANDS", 212),
    Numeric("RPL_STATSCLINE", 213),
    Numeric("RPL_STATSNLINE", 214),
    Numeric("RPL_STATSILINE", 215),
    Numeric("RPL_STATSKLINE", 216),
    Numeric("RPL_STATSQLINE", 217),
    Numeric("RPL_STATSYLINE", 218),
    Numeric("RPL_ENDOFSTATS", 219),
    Numeric("RPL_UMODEIS", 221),
    Numeric("RPL_SERVICEINFO", 231),
    Numeric("RPL_ENDOFSERVICES", 232),
    Numeric("RPL_SERVICE", 233),
    Numeric("RPL_SERVLIST", 234),
    Numeric("RPL_SERVLISTEND", 235),
    Numeric("RPL_STATSVLINE", 240),
    Numeric("RPL_STATSLLINE", 241),
    Numeric("RPL_STATSUPTIME", 242),
    Numeric("RPL_STATSOLINE", 243),
    Numeric("RPL_STATSHLINE", 244),
    Numeric("RPL_STATSPING", 246),
    Numeric("RPL_STATSBLINE", 247),
    Numeric("RPL_STATSDLINE", 250),
    Numeric("RPL_LUSERCLIENT", 251),
    Numeric("RPL_LUSEROP", 252),
    Numeric("RPL_LUSERUNKNOWN", 253),
    Numeric("RPL_LUSERCHANNELS", 254),
    Numeric("RPL_LUSERME", 255),
    Numeric("RPL_ADMINME", 256),
    Numeric("RPL_ADMINLOC1", 257),
    Numeric("RPL_ADMINLOC2", 258),
    Numeric("RPL_ADMINEMAIL", 259),
    Numeric("RPL_TRACELOG", 261),
    Numeric("RPL_TRACEEND", 262),
    Numeric("RPL_TRYAGAIN", 263),
    Numeric("RPL_NONE", 300),
    Numeric("RPL_AWAY", 301),
    Numeric("RPL_USERHOST", 302),
    Numeric("RPL_ISON", 303),
    Numeric("RPL_UNAWAY", 305),
    Numeric("RPL_NOWAWAY", 306),
    Numeric("RPL_WHOISUSER", 311),
    Numeric("RPL_WHOISSERVER", 312),
    Numeric("RPL_WHOISOPERATOR", 313),
    Numeric("RPL_WHOWASUSER", 314),
    Numeric("RPL_ENDOFWHO", 315),
    Numeric("RPL_WHOISCHANOP", 316),
    Numeric("RPL_WHOISIDLE", 317),
    Numeric("RPL_ENDOFWHOIS", 318),
    Numeric("RPL_WHOISCHANNELS", 319),
    Numeric("RPL_LISTSTART", 321),
    Numeric("RPL_LIST", 322),
    Numeric("RPL_LISTEND", 323),
    Numeric("RPL_CHANNELMODEIS", 324),
    Numeric("RPL_UNIQOPIS", 325),
    Numeric("RPL_NOTOPIC", 331),
    Numeric("RPL_TOPIC", 332),
    Numeric("RPL_INVITING", 341),
    Numeric("RPL_SUMMONING", 342),
    Numeric("RPL_INVITELIST", 346),
    Numeric("RPL_ENDOFINVITELIST", 347),
    Numeric("RPL_EXCEPTLIST", 348),
    Numeric("RPL_ENDOFEXCEPTLIST", 349),
    Numeric("RPL_VERSION", 351),
    Numeric("RPL_WHOREPLY", 352),
    Numeric("RPL_NAMREPLY", 353),
    Numeric("RPL_KILLDONE", 361),
    Numeric("RPL_CLOSING", 362),
    Numeric("RPL_CLOSEEND", 363),
    Numeric("RPL_LINKS", 364),
    Numeric("RPL_ENDOFLINKS", 365),
    Numeric("RPL_ENDOFNAMES", 366),
    Numeric("RPL_BANLIST", 367),
    Numeric("RPL_ENDOFBANLIST", 368),
    Numeric("RPL_ENDOFWHOWAS", 369),
    Numeric("RPL_INFO", 371),
    Numeric("RPL_MOTD", 372),
    Numeric("RPL_INFOSTART", 373),
    Numeric("RPL_ENDOFINFO", 374),
    Numeric("RPL_MOTDSTART", 375),
    Numeric("RPL_ENDOFMOTD", 376),
    Numeric("RPL_YOUREOPER", 381),
    Numeric("RPL_REHASHING", 382),
    Numeric("RPL_YOURESERVICE", 383),
    Numeric("RPL_MYPORTIS", 384),
    Numeric("RPL_TIME", 391),
    Numeric("RPL_USERSSTART", 392),
    Numeric("RPL_USERS", 393),
    Numeric("RPL_ENDOFUSERS", 394),
    Numeric("RPL_NOUSERS", 395),
    Numeric("ERR_NOSUCHNICK", 401),
    Numeric("ERR_NOSUCHSERVER", 402),
    Numeric("ERR_NOSUCHCHANNEL", 403),
    Numeric("ERR_CANNOTSENDTOCHAN", 404),
    Numeric("ERR_TOOMANYCHANNELS", 405),
    Numeric("ERR_WASNOSUCHNICK", 406),
    Numeric("ERR_TOOMANYTARGETS", 407),
    Numeric("ERR_NOSUCHSERVICE", 408),
    Numeric("ERR_NOORIGIN", 409),
    Numeric("ERR_NORECIPIENT", 411),
    Numeric("ERR_NOTEXTTOSEND", 412),
    Numeric("ERR_NOTOPLEVEL", 413),
    Numeric("ERR_WILDTOPLEVEL", 414),
    Numeric("ERR_BADMASK", 415),
    Numeric("ERR_UNKNOWNCOMMAND", 421),
    Numeric("ERR_NOMOTD", 422),
    Numeric("ERR_NOADMININFO", 423),
    Numeric("ERR_FILEERROR", 424),
    Numeric("ERR_NONICKNAMEGIVEN", 431),
    Numeric("ERR_ERRONEUSNICKNAME", 432),
    Numeric("ERR_NICKNAMEINUSE", 433),
    Numeric("ERR_NICKCOLLISION", 436),
    Numeric("ERR_UNAVAILRESOURCE", 437),
    Numeric("ERR_USERNOTINCHANNEL", 441),
    Numeric("ERR_NOTONCHANNEL", 442),
    Numeric("ERR_USERONCHANNEL", 443),
    Numeric("ERR_NOLOGIN", 444),
    Numeric("ERR_SUMMONDISABLED", 445),
    Numeric("ERR_USERSDISABLED", 446),
    Numeric("ERR_NOTREGISTERED", 451),
    Numeric("ERR_NEEDMOREPARAMS", 461),
    Numeric("ERR_ALREADYREGISTERED", 462),
    Numeric("ERR_NOPERMFORHOST", 463),
    Numeric("ERR_PASSWDMISMATCH", 464),
    Numeric("ERR_YOUREBANNEDCREEP", 465),
    Numeric("ERR_YOUWILLBEBANNED", 466),
    Numeric("ERR_KEYSET", 467),
    Numeric("ERR_CHANNELISFULL", 471),
    Numeric("ERR_UNKNOWNMODE", 472),
    Numeric("ERR_INVITEONLYCHAN", 473),
    Numeric("ERR_BANNEDFROMCHAN", 474),
    Numeric("ERR_BADCHANNELKEY", 475),
    Numeric("ERR_BADCHANMASK", 476),
    Numeric("ERR_NOCHANMODES", 477),
    Numeric("ERR_BANLISTFULL", 478),
    Numeric("ERR_NOPRIVILEGES", 481),
    Numeric("ERR_CHANOPRIVSNEEDED", 482),
    Numeric("ERR_CANTKILLSERVER", 483),
    Numeric("ERR_RESTRICTED", 484),
    Numeric("ERR_UNIQOPRIVSNEEDED", 485),
    Numeric("ERR_NOOPERHOST", 491),
    Numeric("ERR_NOSERVICEHOST", 492),
    Numeric("ERR_UMODEUNKNOWNFLAG", 501),
    Numeric("ERR_USERSDONTMATCH", 502),
)



"""IRC parser and utility library."""

from irclib._version import (
    __version__,
    __version_tuple__,
    version,
    version_tuple,
)

__all__ = ["version", "version_tuple", "__version__", "__version_tuple__"]



"""IRC string comparison utilities."""

import re
from typing import Final

__all__ = ("match_mask",)

GLOB_MAP: Final = {"?": ".", "*": ".*"}


def match_mask(mask: str, pattern: str) -> bool:
    """Match hostmask patterns in the standard banmask syntax (eg '*!*@host')."""
    re_pattern = ""
    for c in pattern:
        re_pattern += GLOB_MAP.get(c, re.escape(c))

    regex = re.compile(f"^{re_pattern}$")
    return bool(regex.match(mask))



"""IRC string utils."""

import operator
import string
from collections.abc import Callable
from typing import Final, Literal, NamedTuple, Protocol, SupportsIndex

__all__ = ("Casemap", "RFC1459", "STRICT_RFC1459", "ASCII", "String")


class Casemap(NamedTuple):
    """String case-map.

    Represents a mapping of lower to upper case letters
    """

    lower: str
    upper: str

    @property
    def lower_table(self) -> dict[int, int]:
        """The lower->upper translation table."""
        return str.maketrans(self.lower, self.upper)

    @property
    def upper_table(self) -> dict[int, int]:
        """The upper->lower table."""
        return str.maketrans(self.upper, self.lower)


RFC1459: Final = Casemap(
    "".join(map(chr, range(65, 95))), "".join(map(chr, range(97, 127)))
)
STRICT_RFC1459: Final = Casemap(
    "".join(map(chr, range(65, 94))), "".join(map(chr, range(97, 126)))
)
ASCII: Final = Casemap(
    "".join(map(chr, range(65, 91))), "".join(map(chr, range(97, 123)))
)


class TranslateTable(Protocol):
    def __getitem__(self, item: int, /) -> str | int | None:
        raise NotImplementedError


class String(str):
    """Case-insensitive string."""

    __slots__ = ("_casemap",)

    _casemap: Casemap

    def __new__(
        cls, value: str = "", casemap: Casemap | None = None
    ) -> "String":
        """Construct new String and set casemap."""
        o = str.__new__(cls, value)
        o._casemap = casemap or ASCII  # noqa: SLF001
        return o

    def _wrap(self, value: str) -> "String":
        """Convert value to String with matching casemap."""
        return self.__class__.__new__(self.__class__, value, self.casemap)

    def __internal_cmp(
        self, other: object, cmp: Callable[[str, str], bool]
    ) -> tuple[bool, Literal[True]] | tuple[None, Literal[False]]:
        if isinstance(other, String):
            return cmp(str(self.casefold()), str(other.casefold())), True

        if isinstance(other, str):
            return cmp(self, self._wrap(other)), True

        return None, False

    def translate(self, table: TranslateTable) -> "String":
        """Apply translation table to string."""
        return self._wrap(super().translate(table))

    @property
    def _capitalize(self) -> "String":
        return self[:1].upper() + self[1:].lower()

    def capitalize(self) -> "String":
        """Return a capitalized version of the string.

        More specifically, make the first character have upper case and the rest lower case.
        """
        return self._capitalize

    @property
    def _casefold(self) -> "String":
        return self.lower()

    def casefold(self) -> "String":
        """Casefold the string according to the casemap."""
        return self._casefold

    @property
    def _lower(self) -> "String":
        return self.translate(self.casemap.lower_table)

    def lower(self) -> "String":
        """Lowercase the string according to the casemap."""
        return self._lower

    @property
    def _upper(self) -> "String":
        return self.translate(self.casemap.upper_table)

    def upper(self) -> "String":
        """Uppercase the string according to the casemap."""
        return self._upper

    def count(
        self,
        sub: str,
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> int:
        """Count substring occurrences."""
        return str(self.casefold()).count(
            self._wrap(sub).casefold(), start, end
        )

    def startswith(
        self,
        prefix: str | tuple[str, ...],
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> bool:
        """Check if string starts with a prefix."""
        prefix_list: tuple[str, ...]
        prefix_list = (prefix,) if isinstance(prefix, str) else prefix

        mapped_list = tuple(self._wrap(p).casefold() for p in prefix_list)

        return str(self.casefold()).startswith(mapped_list, start, end)

    def endswith(
        self,
        suffix: str | tuple[str, ...],
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> bool:
        """Check if string ends with a suffix."""
        suffix_list: tuple[str, ...]
        suffix_list = (suffix,) if isinstance(suffix, str) else suffix

        mapped_list = tuple(self._wrap(p).casefold() for p in suffix_list)

        return str(self.casefold()).endswith(mapped_list, start, end)

    def find(
        self,
        sub: str,
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> int:
        """Find the substring in string."""
        return str(self.casefold()).find(self._wrap(sub).casefold(), start, end)

    def rfind(
        self,
        sub: str,
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> int:
        """Perform a reverse find."""
        return str(self.casefold()).rfind(
            self._wrap(sub).casefold(), start, end
        )

    def index(
        self,
        sub: str,
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> int:
        """Find the index of the substring."""
        return str(self.casefold()).index(
            self._wrap(sub).casefold(), start, end
        )

    def rindex(
        self,
        sub: str,
        start: SupportsIndex | None = None,
        end: SupportsIndex | None = None,
    ) -> int:
        """Perform a reverse index."""
        return str(self.casefold()).rindex(
            self._wrap(sub).casefold(), start, end
        )

    def partition(self, sep: str) -> tuple["String", "String", "String"]:
        """Partition string on a separator."""
        pos = self.find(sep)
        if pos < 0:
            return self, self._wrap(""), self._wrap("")

        return self[:pos], self[pos : pos + len(sep)], self[pos + len(sep) :]

    def rpartition(self, sep: str) -> tuple["String", "String", "String"]:
        """Reverse partition a string on a separator."""
        pos = self.rfind(sep)
        if pos < 0:
            return self._wrap(""), self._wrap(""), self

        return self[:pos], self[pos : pos + len(sep)], self[pos + len(sep) :]

    def replace(
        self, old: str, new: str, count: SupportsIndex = -1
    ) -> "String":
        """Not currently implemented."""
        raise NotImplementedError

    def strip(self, chars: str | None = None) -> "String":
        """Remove characters from the beginning and end of the string."""
        return self.lstrip(chars).rstrip(chars)

    def lstrip(self, chars: str | None = None) -> "String":
        """Remove characters from the beginning of the string."""
        if chars is None:
            chars = string.whitespace

        chars = self._wrap(chars)

        start = 0
        while start < len(self) and self[start] in chars:
            start += 1

        return self[start:]

    def rstrip(self, chars: str | None = None) -> "String":
        """Remove characters from the end of the string."""
        if chars is None:
            chars = string.whitespace

        chars = self._wrap(chars)

        end = len(self)
        while end > 0 and self[end - 1] in chars:
            end -= 1

        return self[:end]

    @property
    def _swapcase(self) -> "String":
        trans = self.casemap.lower_table.copy()
        trans.update(self.casemap.upper_table)

        return self.translate(trans)

    def swapcase(self) -> "String":
        """Swap lower-case and upper-case characters."""
        return self._swapcase

    def title(self) -> str:
        """Not currently implemented."""
        raise NotImplementedError

    def split(
        self, sep: str | None = None, maxsplit: SupportsIndex = -1
    ) -> list[str]:
        """Not currently implemented."""
        raise NotImplementedError

    def rsplit(
        self, sep: str | None = None, maxsplit: SupportsIndex = -1
    ) -> list[str]:
        """Not currently implemented."""
        raise NotImplementedError

    @property
    def casemap(self) -> Casemap:
        """Casemap associated with this string."""
        return self._casemap

    def __getitem__(self, item: SupportsIndex | slice) -> "String":
        """Get substring."""
        return self._wrap(super().__getitem__(item))

    def __contains__(self, item: object) -> bool:
        """Check if `item` is in string case-insensitively."""
        if not isinstance(item, str):
            return False

        return self._wrap(item).casefold() in str(self.casefold())

    def __add__(self, other: str) -> "String":
        """Concat a string to this one."""
        return self._wrap(str(self) + other)

    def __lt__(self, other: str) -> bool:
        """Compare another string to this one case-insensitively."""
        res = self.__internal_cmp(other, operator.lt)
        if not res[1]:
            return NotImplemented

        return res[0]

    def __le__(self, other: str) -> bool:
        """Compare another string to this one case-insensitively."""
        res = self.__internal_cmp(other, operator.le)
        if not res[1]:
            return NotImplemented

        return res[0]

    def __eq__(self, other: object) -> bool:
        """Compare another string to this one case-insensitively."""
        res = self.__internal_cmp(other, operator.eq)
        if not res[1]:
            return NotImplemented

        return res[0]

    def __ne__(self, other: object) -> bool:
        """Compare another string to this one case-insensitively."""
        res = self.__internal_cmp(other, operator.ne)
        if not res[1]:
            return NotImplemented

        return res[0]

    def __gt__(self, other: str) -> bool:
        """Compare another string to this one case-insensitively."""
        res = self.__internal_cmp(other, operator.gt)
        if not res[1]:
            return NotImplemented

        return res[0]

    def __ge__(self, other: str) -> bool:
        """Compare another string to this one case-insensitively."""
        res = self.__internal_cmp(other, operator.ge)
        if not res[1]:
            return NotImplemented

        return res[0]

    def __hash__(self) -> int:
        """Hash the lowercase string."""
        return hash(str(self.lower()))



"""Frozen Dict."""

from collections.abc import Iterable, Iterator, Mapping
from typing import TypeVar

from typing_extensions import Self

__all__ = ("FrozenDict",)

_V = TypeVar("_V")


class FrozenDict(Mapping[str, _V]):
    """Frozen Mapping.

    An immutable mapping of string -> Any type
    """

    __slots__ = ("__hash", "__data")

    def __init__(
        self,
        seq: Mapping[str, _V] | Iterable[tuple[str, _V]] | None = None,
        **kwargs: _V,
    ) -> None:
        """Construct a FrozenDict."""
        d = dict(seq, **kwargs) if seq is not None else dict(**kwargs)

        self.__data: dict[str, _V] = d
        self.__hash: int | None = None

    def copy(self, **kwargs: _V) -> Self:
        """Copy dict, replacing values according to kwargs.

        >>> fd = FrozenDict(a=1)
        >>> fd["a"]
        1
        >>> fd1 = fd.copy(a=2)
        >>> fd1["a"]
        2
        >>> fd["a"]
        1
        """
        return self.__class__(self.__data, **kwargs)

    def __getitem__(self, k: str) -> _V:
        """Retrieve an item from the dict by key."""
        return self.__data[k]

    def __iter__(self) -> Iterator[str]:
        """Iterate the keys of the dict."""
        return iter(self.__data)

    def __len__(self) -> int:
        """Number of keys in the dict."""
        return len(self.__data)

    def __hash__(self) -> int:
        """Get hash for the dict."""
        if self.__hash is None:
            self.__hash = hash(tuple(self.items()))

        return self.__hash



"""IRC command data and utilities."""

from collections.abc import Iterator, Mapping
from typing import cast

import attr

__all__ = ("Command", "client_commands")


@attr.s(frozen=True, hash=True, auto_attribs=True)
class CommandArgument:
    """A single IRC command argument."""

    name: str
    required: bool | None = True

    @classmethod
    def parse(cls, s: str) -> "CommandArgument":
        """Parse a CommandArgument from an argument string.

        >>> CommandArgument.parse("[foo]")
        CommandArgument(name='foo', required=False)
        >>> CommandArgument.parse("<foo>")
        CommandArgument(name='foo', required=True)

        :param s: String to parse
        :return: Parsed argument
        """
        start_end = s[0] + s[-1]
        name = s[1:-1]
        if start_end == "<>":
            required = True
        elif start_end == "[]":
            required = False
        else:
            msg = f"Unable to parse argument: {s}"
            raise ValueError(msg)

        return cls(name, required)


@attr.s(frozen=True, hash=True, auto_attribs=True)
class Command:
    """A single IRC command."""

    name: str
    args: list[CommandArgument]
    min_args: int = 0
    max_args: int | None = None


class LookupDict(Mapping[str, Command]):
    """Command lookup dictionary."""

    def __init__(self, *commands: Command) -> None:
        for command in commands:
            setattr(self, command.name.lower(), command)
            setattr(self, command.name.upper(), command)

    def __getitem__(self, key: str) -> Command:
        try:
            return cast(Command, getattr(self, key))
        except AttributeError as e:
            raise KeyError(key) from e

    def __iter__(self) -> Iterator[str]:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError


# Commands sent from the client to the server
client_commands = LookupDict(
    Command(
        "PRIVMSG",
        [CommandArgument.parse("<target>"), CommandArgument.parse("<content>")],
    ),
    Command(
        "NOTICE",
        [CommandArgument.parse("<target>"), CommandArgument.parse("<content>")],
    ),
    Command(
        "JOIN",
        [CommandArgument.parse("<channel>"), CommandArgument.parse("[key]")],
    ),
)
